# Example Python plugin.
#
# This plugin simulates errors for testing NBD client error handling.
# Every odd call will fail, and every even call will succeed, unless
# there a real error accessing the specified file.
#
# This example can be freely used for any purpose.

# Run it from the build directory like this:
#
#   ./nbdkit -f -v python ./plugins/python/examples/error.py file=test.img
#
# Or run it after installing nbdkit like this:
#
#   nbdkit -f -v python ./plugins/python/examples/error.py file=test.img
#
# The -f -v arguments are optional.  They cause the server to stay in
# the foreground and print debugging, which is useful when testing.

import os
import nbdkit

API_VERSION = 2

filename = None
calls = 0


def config(key, value):
    global filename
    assert key == "file"
    filename = value


def thread_model():
    # Serialize all requests so we can seek safely in pread and pwrite
    # and be compatible with python 3.6. In python 3.7 we can use
    # os.preadv and os.pwritev and use the parallel threading model.
    return nbdkit.THREAD_MODEL_SERIALIZE_ALL_REQUESTS


def open(readonly):
    flags = os.O_RDONLY if readonly else os.O_RDWR
    return {"fd": os.open(filename, flags)}


def can_extents(h):
    return True


def get_size(h):
    return os.stat(h["fd"]).st_size


def extents(h, count, offset, flags):
    global calls
    calls += 1
    if calls % 2:
        raise RuntimeError(f"extents error offset={offset} count={count}")

    # We don't really support extents, so we report the entire file as
    # data.
    return [(offset, count, 0)]


def pread(h, buf, offset, flags):
    global calls
    calls += 1
    if calls % 2:
        raise RuntimeError(f"pread error offset={offset} count={len(buf)}")

    os.lseek(h["fd"], offset, os.SEEK_SET)
    n = os.readv(h['fd'], [buf])
    assert n == len(buf)


def pwrite(h, buf, offset, flags):
    global calls
    calls += 1
    if calls % 2:
        raise RuntimeError(f"pwrite error offset={offset} count={len(buf)}")

    os.lseek(h["fd"], offset, os.SEEK_SET)
    n = os.writev(h['fd'], [buf])
    assert n == len(buf)



# Example Python plugin.
#
# This example can be freely used for any purpose.
#
# Upload and download images to oVirt with nbdkit and qemu-img.
#
# This assumes you have an oVirt environment. If you don't please see
# https://ovirt.org/ for info on how to install oVirt.
#
# Install ovirt-release package:
#
#  dnf install http://resources.ovirt.org/pub/yum-repo/ovirt-release-master.rpm
#
# Install required packages:
#
#   dnf install ovirt-imageio-client python3-ovirt-engine-sdk4
#
# Note: python3-ovirt-engine-sdk4 is not available yet for Fedora 31 and 32.
#
# To upload or download images, you need to start an image transfer. The
# easiest way is using oVirt image_transfer.py example:
#
#  /usr/share/doc/python3-ovirt-engine-sdk4/examples/image_transfer.py \
#      --engine-url https://my.engine \
#      --username admin@internal \
#      --password-file password \
#      --cafile my.engine.pem \
#      upload disk-uuid
#
# This will print the transfer URL for this image transfer.
#
# Run this example from the build directory:
#
#   ./nbdkit -t4 -f -v -U /tmp/nbd.sock python \
#       ./plugins/python/examples/imageio.py \
#       transfer_url=https://server:54322/images/ticket-id \
#       connections=4 \
#       secure=no
#
# Note that number of nbdkit threads and imageio connections should match.
#
# To upload an image run:
#
#   qemu-img convert -n -f raw -O raw -W disk.img \
#       nbd+unix:///\?socket=/tmp/nbd.sock
#
# Downloading image is not efficient with this version, since we don't report
# extents yet.
#
# The -f -v arguments are optional.  They cause the server to stay in
# the foreground and print debugging, which is useful when testing.

import queue
from contextlib import contextmanager

from ovirt_imageio.client import ImageioClient

import nbdkit

# Using version 2 supporting the buffer protocol for better performance.
API_VERSION = 2

# Plugin configuration, can be set using key=value in the command line.
params = {
    "secure": True,
    "ca_file": "",
    "connections": 1,
    "transfer_url": None,
}


def config(key, value):
    """
    Parse the url parameter which contains the transfer URL that we want to
    serve.
    """
    if key == "transfer_url":
        params["transfer_url"] = value
    elif key == "connections":
        params["connections"] = int(value)
    elif key == "ca_file":
        params["ca_file"] = value
    elif key == "secure":
        params["secure"] = boolify(key, value)
    else:
        raise RuntimeError("unknown parameter: {!r}".format(key))


def boolify(key, value):
    v = value.lower()
    if v in ("yes", "true", "1"):
        return True
    if v in ("no", "false", 0):
        return False
    raise RuntimeError("Invalid boolean value for {}: {!r}".format(key, value))


def config_complete():
    """
    Called when configuration is completed.
    """
    if params["transfer_url"] is None:
        raise RuntimeError("'transfer_url' parameter is required")


def thread_model():
    """
    Using parallel model to speed up transfer with multiple connections to
    imageio server.
    """
    return nbdkit.THREAD_MODEL_PARALLEL


def open(readonly):
    """
    Called once when plugin is loaded. We create a pool of connected clients
    that will be used for requests later.
    """
    pool = queue.Queue()
    for i in range(params["connections"]):
        client = ImageioClient(
            params["transfer_url"],
            cafile=params["ca_file"],
            secure=params["secure"])
        pool.put(client)
    return {"pool": pool}


def close(h):
    """
    Called when plugin is closed. Close and remove all clients from the pool.
    """
    pool = h["pool"]
    while not pool.empty():
        client = pool.get()
        client.close()


@contextmanager
def client(h):
    """
    Context manager fetching an imageio client from the pool. Blocks until a
    client is available.
    """
    pool = h["pool"]
    client = pool.get()
    try:
        yield client
    finally:
        pool.put(client)


def get_size(h):
    with client(h) as c:
        return c.size()


def pread(h, buf, offset, flags):
    with client(h) as c:
        c.read(offset, buf)


def pwrite(h, buf, offset, flags):
    with client(h) as c:
        c.write(offset, buf)


def zero(h, count, offset, flags):
    with client(h) as c:
        c.zero(offset, count)


def flush(h, flags):
    with client(h) as c:
        c.flush()



# nbdkit test plugin
# Copyright Red Hat
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# * Neither the name of Red Hat nor the names of its contributors may be
# used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY RED HAT AND CONTRIBUTORS ''AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL RED HAT OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

"""See tests/test-python.sh."""

import nbdkit
import pickle
import base64
import unittest

API_VERSION = 2

cfg = {}


# Not nice, but there doesn't seem to be a better way of putting this
class TestAPI(unittest.TestCase):

    def test_parse_size(self):
        self.assertEqual(nbdkit.parse_size('511'), 511)
        self.assertEqual(nbdkit.parse_size('7k'), 7*1024)
        self.assertEqual(nbdkit.parse_size('17M'), 17*1024*1024)

        with self.assertRaises(TypeError):
            nbdkit.parse_size(17)

        with self.assertRaises(ValueError):
            nbdkit.parse_size('foo')

    def test_parse_probability(self):
        test = "test_parse_probability"

        self.assertEqual(nbdkit.parse_probability(test, '1:10'), 0.1)
        self.assertEqual(nbdkit.parse_probability(test, '100%'), 1)
        self.assertEqual(nbdkit.parse_probability(test, '0'), 0)

        with self.assertRaises(TypeError):
            nbdkit.parse_probability(test, 17)

        with self.assertRaises(ValueError):
            nbdkit.parse_probability(test, 'bar')

    def test_parse_delay(self):
        test = "test_parse_delay"

        self.assertEqual(nbdkit.parse_delay(test, '10ms'), (0, 10000000))
        self.assertEqual(nbdkit.parse_delay(test, '1.5'), (1, 500000000))

        with self.assertRaises(TypeError):
            nbdkit.parse_delay(test, 17)

        with self.assertRaises(ValueError):
            nbdkit.parse_delay(test, 'bar')

    def test_parse_bool(self):
        self.assertEqual(nbdkit.parse_bool("true"), True)
        self.assertEqual(nbdkit.parse_bool("yes"), True)
        self.assertEqual(nbdkit.parse_bool("no"), False)
        self.assertEqual(nbdkit.parse_bool("n"), False)
        self.assertEqual(nbdkit.parse_bool("0"), False)

        with self.assertRaises(TypeError):
            nbdkit.parse_bool(123)

        with self.assertRaises(ValueError):
            nbdkit.parse_bool("abc")


TestAPI().test_parse_size()
TestAPI().test_parse_probability()
TestAPI().test_parse_delay()
TestAPI().test_parse_bool()


def config(k, v):
    global cfg
    if k == "cfg":
        cfg = pickle.loads(base64.b64decode(v.encode()))


def config_complete():
    print("set_error = %r" % nbdkit.set_error)


def open(readonly):
    if cfg.get('create_disk', True):
        disk = bytearray(cfg.get('size', 0))
    else:
        disk = None
    return {
        'disk': disk
    }


def get_size(h):
    return cfg.get('size', 0)


def is_rotational(h):
    return cfg.get('is_rotational', False)


def can_multi_conn(h):
    return cfg.get('can_multi_conn', False)


def can_write(h):
    return cfg.get('can_write', True)


def can_flush(h):
    return cfg.get('can_flush', False)


def can_trim(h):
    return cfg.get('can_trim', False)


def can_zero(h):
    return cfg.get('can_zero', False)


def can_fast_zero(h):
    return cfg.get('can_fast_zero', False)


def can_fua(h):
    fua = cfg.get('can_fua', "none")
    if fua == "none":
        return nbdkit.FUA_NONE
    elif fua == "emulate":
        return nbdkit.FUA_EMULATE
    elif fua == "native":
        return nbdkit.FUA_NATIVE


def can_cache(h):
    cache = cfg.get('can_cache', "none")
    if cache == "none":
        return nbdkit.CACHE_NONE
    elif cache == "emulate":
        return nbdkit.CACHE_EMULATE
    elif cache == "native":
        return nbdkit.CACHE_NATIVE


def can_extents(h):
    return cfg.get('can_extents', False)


def pread(h, buf, offset, flags):
    assert flags == 0
    end = offset + len(buf)
    buf[:] = h['disk'][offset:end]


def pwrite(h, buf, offset, flags):
    expect_fua = cfg.get('pwrite_expect_fua', False)
    actual_fua = bool(flags & nbdkit.FLAG_FUA)
    assert expect_fua == actual_fua
    end = offset + len(buf)
    assert h['disk'] is not None
    h['disk'][offset:end] = buf


def flush(h, flags):
    assert flags == 0


def trim(h, count, offset, flags):
    expect_fua = cfg.get('trim_expect_fua', False)
    actual_fua = bool(flags & nbdkit.FLAG_FUA)
    assert expect_fua == actual_fua
    if h['disk'] is not None:
        h['disk'][offset:offset+count] = bytearray(count)


def zero(h, count, offset, flags):
    expect_fua = cfg.get('zero_expect_fua', False)
    actual_fua = bool(flags & nbdkit.FLAG_FUA)
    assert expect_fua == actual_fua
    expect_may_trim = cfg.get('zero_expect_may_trim', False)
    actual_may_trim = bool(flags & nbdkit.FLAG_MAY_TRIM)
    assert expect_may_trim == actual_may_trim
    expect_fast_zero = cfg.get('zero_expect_fast_zero', False)
    actual_fast_zero = bool(flags & nbdkit.FLAG_FAST_ZERO)
    assert expect_fast_zero == actual_fast_zero
    if h['disk'] is not None:
        h['disk'][offset:offset+count] = bytearray(count)


def cache(h, count, offset, flags):
    assert flags == 0
    # do nothing


def extents(h, count, offset, flags):
    return cfg.get('extents', [])



#!/usr/bin/env python3
# nbdkit
# Copyright Red Hat
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# * Neither the name of Red Hat nor the names of its contributors may be
# used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY RED HAT AND CONTRIBUTORS ''AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL RED HAT OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

"""
This tests the Python plugin thoroughly by issuing client commands
through libnbd and checking we get the expected results.  It uses an
associated plugin (test-python-plugin.sh).
"""

import nbd
import unittest
import pickle
import base64


class Test(unittest.TestCase):
    def setUp(self):
        self.h = nbd.NBD()

    def tearDown(self):
        del self.h

    def connect(self, cfg):
        cfg = base64.b64encode(pickle.dumps(cfg)).decode()
        cmd = ["nbdkit", "-v", "-s", "--exit-with-parent",
               "python", "test-python-plugin.py", "cfg=" + cfg]
        self.h.connect_command(cmd)

    def test_none(self):
        """
        Test we can send an empty pickled test configuration and do
        nothing else.  This is just to ensure the machinery of the
        test works.
        """
        self.connect({})

    def test_size_512(self):
        """Test the size."""
        self.connect({"size": 512})
        assert self.h.get_size() == 512

    def test_size_1m(self):
        """Test the size."""
        self.connect({"size": 1024*1024})
        assert self.h.get_size() == 1024*1024

    # Test each flag call.
    def test_is_rotational_true(self):
        self.connect({"size": 512, "is_rotational": True})
        assert self.h.is_rotational()

    def test_is_rotational_false(self):
        self.connect({"size": 512, "is_rotational": False})
        assert not self.h.is_rotational()

    def test_can_multi_conn_true(self):
        self.connect({"size": 512, "can_multi_conn": True})
        assert self.h.can_multi_conn()

    def test_can_multi_conn_false(self):
        self.connect({"size": 512, "can_multi_conn": False})
        assert not self.h.can_multi_conn()

    def test_read_write(self):
        self.connect({"size": 512, "can_write": True})
        assert not self.h.is_read_only()

    def test_read_only(self):
        self.connect({"size": 512, "can_write": False})
        assert self.h.is_read_only()

    def test_can_flush_true(self):
        self.connect({"size": 512, "can_flush": True})
        assert self.h.can_flush()

    def test_can_flush_false(self):
        self.connect({"size": 512, "can_flush": False})
        assert not self.h.can_flush()

    def test_can_trim_true(self):
        self.connect({"size": 512, "can_trim": True})
        assert self.h.can_trim()

    def test_can_trim_false(self):
        self.connect({"size": 512, "can_trim": False})
        assert not self.h.can_trim()

    # nbdkit can always zero because it emulates it.
    # self.connect({"size": 512, "can_zero": True})
    # assert self.h.can_zero()
    # self.connect({"size": 512, "can_zero": False})
    # assert not self.h.can_zero()

    def test_can_fast_zero_true(self):
        self.connect({"size": 512, "can_fast_zero": True})
        assert self.h.can_fast_zero()

    def test_can_fast_zero_false(self):
        self.connect({"size": 512, "can_fast_zero": False})
        assert not self.h.can_fast_zero()

    def test_can_fua_none(self):
        self.connect({"size": 512, "can_fua": "none"})
        assert not self.h.can_fua()

    def test_can_fua_emulate(self):
        self.connect({"size": 512, "can_fua": "emulate"})
        assert self.h.can_fua()

    def test_can_fua_native(self):
        self.connect({"size": 512, "can_fua": "native"})
        assert self.h.can_fua()

    def test_can_cache_none(self):
        self.connect({"size": 512, "can_cache": "none"})
        assert not self.h.can_cache()

    def test_can_cache_emulate(self):
        self.connect({"size": 512, "can_cache": "emulate"})
        assert self.h.can_cache()

    def test_can_cache_native(self):
        self.connect({"size": 512, "can_cache": "native"})
        assert self.h.can_cache()

    # In theory we could use a test like this, but nbdkit can
    # always synthesize base:allocation block_status responses
    # even if the plugin doesn't support them.
    #
    # def test_can_extents_true(self):
    #     self.h.add_meta_context("base:allocation")
    #     self.connect({"size": 512, "can_extents": True})
    #     assert self.h.can_meta_context("base:allocation")
    #
    # def test_can_extents_false(self):
    #     self.h.add_meta_context("base:allocation")
    #     self.connect({"size": 512, "can_extents": False})
    #     assert not self.h.can_meta_context("base:allocation")

    def test_pread(self):
        """Test pread."""
        self.connect({"size": 512})
        buf = self.h.pread(512, 0)
        assert buf == bytearray(512)

    # Test pwrite + flags.
    def test_pwrite(self):
        self.connect({"size": 512})
        buf = bytearray(512)
        self.h.pwrite(buf, 0)

    def test_pwrite_fua(self):
        self.connect({"size": 512,
                      "can_fua": "native",
                      "pwrite_expect_fua": True})
        buf = bytearray(512)
        self.h.pwrite(buf, 0, nbd.CMD_FLAG_FUA)

    def test_flush(self):
        """Test flush."""
        self.connect({"size": 512, "can_flush": True})
        self.h.flush()

    def test_trim(self):
        """Test trim."""
        self.connect({"size": 512, "can_trim": True})
        self.h.trim(512, 0)

    def test_trim_fua(self):
        """Test trim + flags."""
        self.connect({"size": 512,
                      "can_trim": True,
                      "can_fua": "native",
                      "trim_expect_fua": True})
        self.h.trim(512, 0, nbd.CMD_FLAG_FUA)

    def test_zero(self):
        """Test zero."""
        self.connect({"size": 512, "can_zero": True})
        self.h.zero(512, 0, nbd.CMD_FLAG_NO_HOLE)

    def test_zero_fua(self):
        """Test zero + flags."""
        self.connect({"size": 512,
                      "can_zero": True,
                      "can_fua": "native",
                      "zero_expect_fua": True})
        self.h.zero(512, 0, nbd.CMD_FLAG_NO_HOLE | nbd.CMD_FLAG_FUA)

    def test_zero_may_trim(self):
        """Test zero + may trim."""
        self.connect({"size": 512,
                      "can_zero": True,
                      "zero_expect_may_trim": True})
        self.h.zero(512, 0, 0)  # absence of nbd.CMD_FLAG_NO_HOLE

    def test_zero_fast_zero(self):
        """Test zero + fast zero."""
        self.connect({"size": 512,
                      "can_zero": True,
                      "can_fast_zero": True,
                      "zero_expect_fast_zero": True})
        self.h.zero(512, 0, nbd.CMD_FLAG_NO_HOLE | nbd.CMD_FLAG_FAST_ZERO)

    def test_zero_large(self):
        """Test large zero."""
        self.connect({"size": 0x10_0000_0000,
                      "can_zero": True,
                      "create_disk": False})
        self.h.zero(0xf000_0000, 0, nbd.CMD_FLAG_NO_HOLE)

    def test_cache(self):
        """Test cache."""
        self.connect({"size": 512, "can_cache": "native"})
        self.h.cache(512, 0)

    # We don't have access to the magic constants defined in the
    # nbdkit module, so redefine them here.
    EXTENT_HOLE = 1
    EXTENT_ZERO = 2

    def test_extents_1(self):
        """Test extents."""

        offset = None
        entries = []

        def f(meta_context, o, e, err):
            nonlocal offset, entries
            if meta_context != "base:allocation":
                return
            offset = o
            entries = e

        self.h.add_meta_context("base:allocation")
        self.connect({"size": 512,
                      "can_extents": True,
                      "extents":
                      [(0, 512, self.EXTENT_HOLE | self.EXTENT_ZERO)]})

        self.h.block_status(512, 0, f)
        assert offset == 0
        assert entries == [512, self.EXTENT_HOLE | self.EXTENT_ZERO]

    def test_extents_2(self):
        """Test extents."""

        offset = None
        entries = []

        def f(meta_context, o, e, err):
            nonlocal offset, entries
            if meta_context != "base:allocation":
                return
            offset = o
            entries = e

        self.h.add_meta_context("base:allocation")
        self.connect({"size": 2048,
                      "can_extents": True,
                      "extents":
                      [(0, 512, self.EXTENT_HOLE | self.EXTENT_ZERO),
                       (512, 512, 0),
                       (1024, 1024, self.EXTENT_ZERO)]})

        self.h.block_status(2048, 0, f)
        assert offset == 0
        assert entries == [512, self.EXTENT_HOLE | self.EXTENT_ZERO,
                           512, 0,
                           1024, self.EXTENT_ZERO]

        self.h.block_status(1024, 1024, f)
        assert offset == 1024
        assert entries == [1024, self.EXTENT_ZERO]



# Example Python plugin.
#
# This example can be freely used for any purpose.

# Run it from the build directory like this:
#
#   ./nbdkit -f -v python ./plugins/python/examples/file.py file=test.img
#
# Or run it after installing nbdkit like this:
#
#   nbdkit -f -v python ./plugins/python/examples/file.py file=test.img
#
# The -f -v arguments are optional.  They cause the server to stay in
# the foreground and print debugging, which is useful when testing.

import os

import nbdkit

# There are several variants of the API.  nbdkit will use this
# constant to determine which one you want to use.  This is the latest
# version at the time this example was written.
API_VERSION = 2

# The file we want to serve.
filename = None


# Parse the file parameter which contains the name of the file that we
# want to serve.
def config(key, value):
    global filename
    if key == "file":
        filename = os.path.abspath(value)
    else:
        raise RuntimeError("unknown parameter: " + key)


def config_complete():
    if filename is None:
        raise RuntimeError("file parameter is required")


# Although Python code cannot be run in parallel, if your
# plugin callbacks sleep then you can improve parallelism
# by relaxing the thread model.
def thread_model():
    return nbdkit.THREAD_MODEL_PARALLEL


# This is called when a client connects.
def open(readonly):
    if readonly:
        flags = os.O_RDONLY
    else:
        flags = os.O_RDWR
    fd = os.open(filename, flags)
    return {'fd': fd}


def get_size(h):
    sb = os.stat(h['fd'])
    return sb.st_size


def pread(h, buf, offset, flags):
    n = os.preadv(h['fd'], [buf], offset)
    if n != len(buf):
        raise RuntimeError("unexpected short read from file")


def pwrite(h, buf, offset, flags):
    n = os.pwritev(h['fd'], [buf], offset)
    if n != len(buf):
        raise RuntimeError("short write")



# nbdkit
# Copyright Red Hat
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# * Neither the name of Red Hat nor the names of its contributors may be
# used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY RED HAT AND CONTRIBUTORS ''AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL RED HAT OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

# Create an nbdkit python plugin which checks the peer UID & GID.

import nbdkit


def config(key, value):
    global pid, uid, gid
    if key == "pid":
        pid = int(value)
    elif key == "uid":
        uid = int(value)
    elif key == "gid":
        gid = int(value)
    else:
        raise RuntimeError("unknown parameter: " + key)


def open(readonly):
    # Fail open() if either we cannot read the peer credentials or
    # they do not match the parameters.
    p = nbdkit.peer_pid()
    nbdkit.debug("nbdkit.peer_pid() = %d" % p)
    u = nbdkit.peer_uid()
    nbdkit.debug("nbdkit.peer_uid() = %d" % u)
    g = nbdkit.peer_gid()
    nbdkit.debug("nbdkit.peer_gid() = %d" % g)

    assert p == pid
    assert u == uid
    assert g == gid

    return 1


def get_size(h):
    return 0


def pread(h, count, offset):
    return bytearray(count)



# Example Python plugin.
#
# This example can be freely used for any purpose.

# Run it from the build directory like this:
#
#   ./nbdkit -f -v python ./plugins/python/examples/url.py \
#       url=http://example.com/disk.img
#
# Or run it after installing nbdkit like this:
#
#   nbdkit -f -v python ./plugins/python/examples/url.py \
#       url=http://example.com/disk.img
#
# The -f -v arguments are optional.  They cause the server to stay in
# the foreground and print debugging, which is useful when testing.

import urllib.request

import nbdkit

# There are several variants of the API.  nbdkit will use this
# constant to determine which one you want to use.  This is the latest
# version at the time this example was written.
API_VERSION = 2

url = None


# Parse the url parameter.
def config(key, value):
    global url
    if key == "url":
        url = value
    else:
        raise RuntimeError("unknown parameter: " + key)


def config_complete():
    if url is None:
        raise RuntimeError("url parameter is required")


# Although Python code cannot be run in parallel, if your
# plugin callbacks sleep then you can improve parallelism
# by relaxing the thread model.
def thread_model():
    return nbdkit.THREAD_MODEL_PARALLEL


# This is called when a client connects.
def open(readonly):
    return 1


def get_size(h):
    rq = urllib.request.Request(url, method='HEAD')
    rp = urllib.request.urlopen(rq)
    headers = rp.info()

    # Check the server supports range requests.
    if headers.get_all('accept-ranges') == []:
        raise RuntimeError("server does not support range requests")

    content_length = int(headers.get_all('content-length')[0])
    return content_length


def pread(h, buf, offset, flags):
    count = len(buf)
    headers = {"range": "bytes=%d-%d" % (offset, offset+count-1)}
    rq = urllib.request.Request(url, b"", headers)
    rp = urllib.request.urlopen(rq)
    b = rp.read()
    if len(b) != count:
        raise RuntimeError("incorrect length of data returned: perhaps the " +
                           "server does not really support range requests")
    buf[:] = b



#!@sbindir@/nbdkit python
# -*- python -*-
# Copyright Infrascale Inc.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# * Neither the name of Infrascale nor the names of its contributors may be
# used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY INFRASCALE AND CONTRIBUTORS ''AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INFRASCALE OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

import base64
import errno
import os
import secrets
import tempfile
import threading
import unittest
from contextlib import contextmanager
from typing import Optional, Union, List
from unittest.mock import patch

import nbdkit

from google.api_core.exceptions import NotFound, GatewayTimeout, \
    DeadlineExceeded
from google.cloud import storage as gcs


nbdkit.debug("gcs: imported google.cloud.storage from %s" % gcs.__file__)

API_VERSION = 2
BinaryData = Union[bytes, bytearray, memoryview]


class Config:
    """Holds configuration data passed in by the user."""

    def __init__(self) -> None:
        self.json_creds = None
        self.bucket_name = None
        self.key_name = None
        self.dev_size = None
        self.obj_size = None

    def set(self, key: str, value: str) -> None:
        """Set a configuration value."""

        if key == "json-credentials" or key == "json_credentials":
            self.json_creds = value
        elif key == "bucket":
            self.bucket_name = value
        elif key == "key":
            self.key_name = value
        elif key == 'size':
            self.dev_size = nbdkit.parse_size(value)
        elif key == 'object-size' or key == "object_size":
            self.obj_size = nbdkit.parse_size(value)
        else:
            raise RuntimeError("unknown parameter %s" % key)

    def validate(self) -> None:
        """Validate configuration settings."""

        if self.bucket_name is None:
            raise RuntimeError("bucket parameter missing")
        if self.key_name is None:
            raise RuntimeError("key parameter missing")

        if (self.dev_size and not self.obj_size or
                self.obj_size and not self.dev_size):
            raise RuntimeError(
                "`size` and `object-size` parameters must always be "
                "specified together")

        if self.dev_size and self.dev_size % self.obj_size != 0:
            raise RuntimeError('`size` must be a multiple of `object-size`')


def thread_model():
    return nbdkit.THREAD_MODEL_PARALLEL


def can_write(server):
    return cfg.obj_size is not None


def can_multi_conn(server):
    return True


def can_trim(server):
    return True


def can_zero(server):
    return True


def can_fast_zero(server):
    return True


def can_cache(server):
    return nbdkit.CACHE_NONE


def block_size(server):
    if not cfg.obj_size:
        # More or less arbitrary.
        return (1, 512 * 1024, 0xffffffff)

    # Should return (minimum, preferred, maximum) block size. We return the
    # same value for preferred and maximum because even though the plugin can
    # handle arbitrary large and small blocks, the performance penalty is
    # huge, and it is always preferable for the client to split up requests
    # as needed. At the same time, the minimum accessible block size is still
    # 1 byte.
    return (1, cfg.obj_size, cfg.obj_size)


def can_fua(server):
    return nbdkit.FUA_NATIVE


def can_flush(server):
    return True


def config(key, value):
    cfg.set(key, value)


def config_complete():
    cfg.validate()


def open(readonly):
    return Server()


def get_size(server):
    return server.get_size()


def pread(server, buf, offset, flags):
    try:
        return server.pread(buf, offset, flags)
    except (GatewayTimeout, DeadlineExceeded):
        nbdkit.debug('GCS connection timed out on pread()')
        nbdkit.set_error(errno.ETIMEDOUT)


def pwrite(server, buf, offset, flags):
    try:
        return server.pwrite(buf, offset, flags)
    except (GatewayTimeout, DeadlineExceeded):
        nbdkit.debug('GCS connection timed out on write()')
        nbdkit.set_error(errno.ETIMEDOUT)


def trim(server, size, offset, flags):
    try:
        return server.trim(size, offset, flags)
    except (GatewayTimeout, DeadlineExceeded):
        nbdkit.debug('GCS connection timed out on trim()')
        nbdkit.set_error(errno.ETIMEDOUT)


def zero(server, size, offset, flags):
    try:
        return server.zero(size, offset, flags)
    except (GatewayTimeout, DeadlineExceeded):
        nbdkit.debug('GCS connection timed out on trim()')
        nbdkit.set_error(errno.ETIMEDOUT)


def flush(server, flags):
    # Flush is implicitly done on every request.
    return


class Server:
    """Handles NBD requests for one connection."""

    def __init__(self) -> None:
        if cfg.json_creds is None:
            nbdkit.debug(
                'Init GCS client using Application Default Credentials')
            self.gcs = gcs.Client()
        else:
            self.gcs = gcs.Client.from_service_account_json(cfg.json_creds)

        self.bucket = self.gcs.bucket(cfg.bucket_name)

    def get_size(self) -> int:
        if cfg.dev_size:
            return cfg.dev_size

        blob = self.bucket.get_blob(cfg.key_name)
        return 0 if blob is None else blob.size

    def pread(self, buf: Union[bytearray, memoryview],
              offset: int, flags: int) -> None:
        to_read = len(buf)
        if not cfg.obj_size:
            buf[:] = self._get_object(cfg.key_name, size=to_read, off=offset)
            return

        read = 0

        (blockno, block_offset) = divmod(offset, cfg.obj_size)
        while to_read > 0:
            key = f"{cfg.key_name}/{blockno:016x}"
            len_ = min(to_read, cfg.obj_size - block_offset)
            buf[read:read + len_] = self._get_object(
                key, size=len_, off=block_offset
            )
            to_read -= len_
            read += len_
            blockno += 1
            block_offset = 0

    def _get_object(self, obj_name: str, size: int, off: int) -> bytes:
        """Read *size* bytes from *obj_name*, starting at *off*."""

        try:
            blob = self.bucket.blob(obj_name)
            buf = blob.download_as_bytes(start=off, end=off + size - 1,
                                         checksum=None, raw_download=True)
        except NotFound:
            return bytearray(size)

        assert len(buf) == size, f'requested {size} bytes, got {len(buf)}'
        return buf

    def pwrite(self, buf: BinaryData, offset: int, flags: int) -> None:
        # We can ignore FUA flags, because every write is always flushed

        if not cfg.obj_size:
            raise RuntimeError('Unable to write in single-object mode')

        # memoryviews can be sliced without copying the data
        if not isinstance(buf, memoryview):
            buf = memoryview(buf)

        # Calculate block number and offset within the block for the
        # first byte that we need to write.
        (blockno1, block_offset1) = divmod(offset, cfg.obj_size)

        # Calculate block number of the last block that we have to
        # write to, and how many bytes we need to write into it.
        (blockno2, block_len2) = divmod(offset + len(buf), cfg.obj_size)

        # Special case: start and end is within the same one block
        if blockno1 == blockno2 and (block_offset1 != 0 or block_len2 != 0):
            nbdkit.debug(f"pwrite(): write at offset {offset} not aligned, "
                         f"covers bytes {block_offset1} to {block_len2} of "
                         f"block {blockno1}. Rewriting full block...")

            # We could separately fetch the prefix and suffix, but give that
            # we're always writing full blocks, it's likely that the latency of
            # two separate read requests would be much bigger than the savings
            # in volume.
            key = f'{cfg.key_name}/{blockno1:016x}'
            with obj_lock(key):
                fbuf = bytearray(self._get_object(
                    key, size=cfg.obj_size, off=0))
                fbuf[block_offset1:block_len2] = buf
                self._put_object(key, fbuf)
                return

        # First write is not aligned to first block
        if block_offset1:
            nbdkit.debug(
                f"pwrite(): write at offset {offset} not aligned, "
                f"starts {block_offset1} bytes into block {blockno1}. "
                "Rewriting full block...")
            key = f'{cfg.key_name}/{blockno1:016x}'
            with obj_lock(key):
                pre = self._get_object(key, size=block_offset1, off=0)
                len_ = cfg.obj_size - block_offset1
                self._put_object(key, pre + buf[:len_])
                buf = buf[len_:]
                blockno1 += 1

        # Last write is not a full block
        if block_len2:
            nbdkit.debug(f"pwrite(): write at offset {offset} not aligned, "
                         f"ends {cfg.obj_size - block_len2} bytes before "
                         f"block {blockno2 + 1}. Rewriting full block...")
            key = f'{cfg.key_name}/{blockno2:016x}'
            with obj_lock(key):
                len_ = cfg.obj_size - block_len2
                post = self._get_object(key, size=len_, off=block_len2)
                self._put_object(key, self._concat(buf[-block_len2:], post))
                buf = buf[:-block_len2]

        off = 0
        for blockno in range(blockno1, blockno2):
            key = f"{cfg.key_name}/{blockno:016x}"
            with obj_lock(key):
                nbdkit.debug(f"pwrite(): writing block {blockno}...")
                self._put_object(key, buf[off:off + cfg.obj_size])
                off += cfg.obj_size

    def _put_object(self, obj_name: str, buf: BinaryData) -> None:
        """Write *buf* into *obj_name"""

        assert len(buf) == cfg.obj_size

        # GCS library does not support reading from memoryviews :-(
        if isinstance(buf, memoryview):
            buf = buf.tobytes()
        elif isinstance(buf, bytearray):
            buf = bytes(buf)

        blob = self.bucket.blob(obj_name)
        blob.upload_from_string(buf, content_type='application/octet-stream')

    def zero(self, size: int, offset: int, flags: int) -> None:
        nbdkit.debug(f'Processing zero(size={size}, off={offset})')
        if size == 0:
            return

        # Calculate block number and offset within the block for the
        # first byte that we need to write.
        (blockno1, block_offset1) = divmod(offset, cfg.obj_size)

        # Calculate block number of the last block that we have to
        # write to, and how many bytes we need to write into it.
        (blockno2, block_len2) = divmod(offset + size, cfg.obj_size)

        if blockno1 == blockno2:
            nbdkit.debug(f'Zeroing {size} bytes in block {blockno1} '
                         f'(offset {offset})')
            self.pwrite(bytearray(size), offset=offset, flags=0)
            return

        if block_offset1:
            len_ = cfg.obj_size - block_offset1
            nbdkit.debug(f'Zeroing last {len_} bytes of block {blockno1} '
                         f'(offset {offset})')
            self.pwrite(bytearray(len_), offset=offset, flags=0)
            blockno1 += 1

        if block_len2:
            off = cfg.obj_size * blockno2
            nbdkit.debug(f'Zeroing first {block_len2 - 1} bytes of block '
                         f'{blockno2} (offset {off})')
            self.pwrite(bytearray(block_len2), offset=off, flags=0)

        self._delete_objects(blockno1, blockno2)

    def trim(self, size: int, offset: int, flags: int) -> None:
        nbdkit.debug(f'Processing trim(size={size}, off={offset})')
        if size == 0:
            return

        # Note, NBD protocol allows to round the offset up and size down to
        # meet internal alignment constraints.

        # Calculate block number and offset within the block for the
        # first byte that we need to trim.
        (blockno1, block_offset1) = divmod(offset, cfg.obj_size)
        if block_offset1 != 0:
            blockno1 += 1

        # Calculate block number of the block following the last one that we
        # have to trim fully.
        (blockno2, _) = divmod(offset + size, cfg.obj_size)

        if blockno1 == blockno2:
            nbdkit.debug('nothing to delete')
            return

        self._delete_objects(blockno1, blockno2)

    @staticmethod
    def _delete_object_callback(obj: str) -> None:
        """ Called by the library on attempt to delete non-existent object """
        # Simply ignore callback

    def _delete_objects(self, first: int, last: int) -> None:
        """Delete objects *first* (inclusive) to *last* (exclusive)"""
        nbdkit.debug(
            f'Deleting objects {first} (inclusive) to {last} (exclusive)...')

        if first >= last:
            return

        first_key = f"{cfg.key_name}/{first:016x}"
        end_key = f"{cfg.key_name}/{last:016x}"

        to_delete = []
        keys = self._list_objects(f"{cfg.key_name}/", first_key, end_key)
        for key in keys:
            if key >= end_key:
                break

            to_delete.append(key)
            nbdkit.debug(f'Marking object {key} for removal')
            if len(to_delete) >= 1000:
                self.bucket.delete_blobs(to_delete,
                                         on_error=self._delete_object_callback)

                del to_delete[:]

        if to_delete:
            self.bucket.delete_blobs(to_delete,
                                     on_error=self._delete_object_callback)

    def _list_objects(self, prefix: str,
                      first_offset: Optional[str] = None,
                      end_offset: Optional[str] = None) -> List[str]:
        """Return keys for objects in bucket.

        Lists all keys starting with *prefix* in lexicographic order, starting
        with the key following *first_offset* (inclusive) and ending with
        *end_offset* (exclusive).
        """

        blobs = self.bucket.list_blobs(prefix=prefix,
                                       fields="items(name),nextPageToken",
                                       start_offset=first_offset,
                                       end_offset=end_offset)
        for blob in blobs:
            yield blob.name

    @staticmethod
    def _concat(b1, b2):
        """Concatenate two byte-like objects (including memoryviews)"""

        l1 = len(b1)
        l3 = l1 + len(b2)
        b3 = bytearray(l3)
        b3[:l1] = b1
        b3[l1:] = b2

        return b3


class MultiLock:
    """Provides locking for large amounts of entities.

    This class provides locking for a dynamically changing and potentially
    large set of entities, avoiding the need to allocate a separate lock for
    each entity. The `acquire` and `release` methods have an additional
    argument, the locking key, and only locks with the same key can see each
    other (ie, several threads can hold locks with different locking keys at
    the same time).
    """

    def __init__(self):
        self.locked_keys = set()
        self.cond = threading.Condition()

    @contextmanager
    def __call__(self, key):
        self.acquire(key)
        try:
            yield
        finally:
            self.release(key)

    def acquire(self, key):
        """Acquire lock for given key."""

        with self.cond:
            while key in self.locked_keys:
                self.cond.wait()

            self.locked_keys.add(key)

    def release(self, key):
        """Release lock on given key"""

        with self.cond:
            self.locked_keys.remove(key)
            self.cond.notify_all()


###################
# Global state    #
###################
cfg = Config()
obj_lock = MultiLock()


######################
# Unit Tests         #
######################

class BlobWithName:
    def __init__(self, name: str):
        self.name = name


class MockGcsClient:
    def __init__(self):
        self.keys = {}

    def bucket(self, bucket_name: str):
        return MockGcsBucket(self, bucket_name)

    def put_object(self, bucket: str, key: str, content):
        self.keys[(bucket, key)] = bytes(content)

    def get_object(self, bucket: str, key: str, start: int, end: int):
        our_key = (bucket, key)
        if our_key not in self.keys:
            raise NotFound(key)

        if start is not None and end is not None:
            buf = self.keys[our_key][start:end + 1]
        else:
            buf = self.keys[our_key]
        return buf

    def delete_objects(self, bucket: str, blobs, on_error):
        for el in blobs:
            key = (bucket, el)
            if key not in self.keys:
                on_error(el)
            del self.keys[key]

    def list_objects(self, bucket: str, prefix: str,
                     start_offset: Optional[str] = None,
                     end_offset: Optional[str] = None):
        all_keys = sorted(x[1] for x in self.keys
                          if x[0] == bucket and x[1].startswith(prefix))
        obj_list = []
        for k in all_keys:
            if start_offset is not None and k < start_offset:
                continue
            if end_offset is not None and k >= end_offset:
                continue
            obj_list.append(BlobWithName(k))

        return obj_list


class MockGcsBucket:
    def __init__(self, client: MockGcsClient, bucket_name: str):
        self.client = client
        self.bucket_name = bucket_name

    def blob(self, obj_name: str):
        return MockGcsBlob(self, obj_name)

    def list_blobs(self, prefix: str, fields: str,
                   start_offset: Optional[str] = None,
                   end_offset: Optional[str] = None):
        return self.client.list_objects(self.bucket_name, prefix,
                                        start_offset, end_offset)

    def delete_blobs(self, blobs, on_error):
        self.client.delete_objects(self.bucket_name, blobs, on_error)


class MockGcsBlob:
    def __init__(self, bucket: MockGcsBucket, blob_name: str):
        self.bucket = bucket
        self.blob_name = blob_name

    def upload_from_string(self, data, content_type: str):
        self.bucket.client.put_object(self.bucket.bucket_name,
                                      self.blob_name, data)

    def download_as_bytes(self, start: int, end: int, raw_download=True,
                          checksum=None):
        return self.bucket.client.get_object(self.bucket.bucket_name,
                                             self.blob_name, start, end)


class LocalTest(unittest.TestCase):
    def setUp(self):
        super().setUp()

        self.obj_size = 16
        self.dev_size = 20 * self.obj_size
        self.ref_fh = tempfile.TemporaryFile()
        self.ref_fh.truncate(cfg.dev_size)

        config('bucket', 'testbucket')
        config('key', 'nbdkit_test')
        config('object-size', str(self.obj_size))
        config('size', str(self.dev_size))

        config_complete()
        with patch.object(gcs, 'Client') as mock_client:
            mock_client.return_value = MockGcsClient()
            self.gcs = open(False)

    def tearDown(self) -> None:
        super().tearDown()
        self.ref_fh.close()

    @staticmethod
    def get_data(len):
        buf = secrets.token_bytes(len // 2 + 1)
        return base64.b16encode(buf)[:len]

    def compare_to_ref(self):
        fh = self.ref_fh
        bl = self.obj_size
        buf = bytearray(bl)
        fh.seek(0)
        for off in range(0, self.dev_size, bl):
            ref = fh.read(bl)
            pread(self.gcs, buf, off, flags=0)
            self.assertEqual(
                ref, buf, f'mismatch at off={off} (blk {off // bl})')

    def test_write_memoryview(self):
        buf = memoryview(bytearray(cfg.obj_size))
        pwrite(self.gcs, buf, 0, 0)
        pwrite(self.gcs, buf[10:], 42, 0)

    def test_read(self):
        fh = self.ref_fh
        bl = self.obj_size

        # Fill disk
        fh.seek(0)
        for off in range(0, self.dev_size, self.obj_size):
            buf = self.get_data(bl)
            pwrite(self.gcs, buf, offset=off, flags=0)
            fh.write(buf)
        self.compare_to_ref()

        # Test different kinds of read requests
        corner_cases = (
            1, 2,
            bl - 2, bl - 1, bl + 2,
            2 * bl - 1, 2 * bl, 2 * bl + 1,
            5 * bl - 5, 5 * bl, 5 * bl + 5)
        for off in (0,) + corner_cases:
            for len_ in corner_cases:
                buf = bytearray(len_)
                pread(self.gcs, buf, off, flags=0)
                fh.seek(off)
                ref = fh.read(len_)
                self.assertEqual(buf, ref)

    def test_write(self):
        fh = self.ref_fh
        bl = self.obj_size

        # Fill disk
        fh.seek(0)
        for off in range(0, self.dev_size, bl):
            buf = self.get_data(bl)
            pwrite(self.gcs, buf, offset=off, flags=0)
            fh.write(buf)
        self.compare_to_ref()

        # Test different kinds of write requests
        corner_cases = (
            1, 2,
            bl - 2, bl - 1, bl + 2,
            2 * bl - 1, 2 * bl, 2 * bl + 1,
            5 * bl - 5, 5 * bl, 5 * bl + 5)
        for off in (0,) + corner_cases:
            for len_ in corner_cases:
                buf = self.get_data(len_)
                pwrite(self.gcs, buf, off, flags=0)
                fh.seek(off)
                fh.write(buf)
                self.compare_to_ref()

    def test_zero(self):
        fh = self.ref_fh
        bl = self.obj_size

        # Fill disk
        fh.seek(0)
        for off in range(0, self.dev_size, bl):
            buf = self.get_data(bl)
            pwrite(self.gcs, buf, offset=off, flags=0)
            fh.write(buf)
        self.compare_to_ref()

        # Test different kinds of zero requests
        corner_cases = (
            1, 2,
            bl - 2, bl - 1, bl, bl + 2,
            2 * bl - 1, 2 * bl, 2 * bl + 1,
            5 * bl - 5, 5 * bl, 5 * bl + 5)
        for flags in (0, nbdkit.FLAG_MAY_TRIM):
            for off in (0,) + corner_cases:
                for len_ in corner_cases:
                    zero(self.gcs, len_, off, flags=flags)
                    fh.seek(off)
                    fh.write(bytearray(len_))
                    self.compare_to_ref()

                    # Re-fill with data
                    buf = self.get_data(len_)
                    pwrite(self.gcs, buf, off, flags=0)
                    fh.seek(off)
                    fh.write(buf)

    def test_trim(self):
        bl = self.obj_size

        # Fill disk
        for off in range(0, self.dev_size, self.obj_size):
            pwrite(self.gcs, self.get_data(bl), offset=off, flags=0)

        # Test different kinds of trim requests
        corner_cases = (
            1, 2,
            bl - 2, bl - 1, bl + 2,
            2 * bl - 1, 2 * bl, 2 * bl + 1,
            5 * bl - 5, 5 * bl, 5 * bl + 5)
        for off in (0,) + corner_cases:
            for len_ in corner_cases:
                (b1, o1) = divmod(off, bl)
                (b2, o2) = divmod(off + len_, bl)

                obj_count1 = len(list(self.gcs._list_objects(cfg.key_name)))
                trim(self.gcs, len_, off, flags=0)
                obj_count2 = len(list(self.gcs._list_objects(cfg.key_name)))

                blocks_to_delete = b2 - b1
                if o1 != 0 and blocks_to_delete >= 1:
                    blocks_to_delete -= 1

                self.assertEqual(obj_count1 - blocks_to_delete, obj_count2)

                # Re-fill with data
                pwrite(self.gcs, self.get_data(len_), off, flags=0)


# To run unit tests against a real GCS bucket, set the TEST_BUCKET and
# TEST_JSON_CREDENTIALS environment variables. The point of these tests is
# to make sure we're calling GCS correctly, not to test any plugin code.
@unittest.skipIf('TEST_BUCKET' not in os.environ,
                 'TEST_BUCKET environment variable not defined.')
@unittest.skipIf('TEST_JSON_CREDENTIALS' not in os.environ,
                 'TEST_JSON_CREDENTIALS environment variable not '
                 'defined.')
class RemoteTest(unittest.TestCase):
    def setUp(self):
        super().setUp()

        self.obj_size = 64
        self.dev_size = 20 * self.obj_size

        config('bucket', os.environ['TEST_BUCKET'])
        config('json-credentials', os.environ['TEST_JSON_CREDENTIALS'])
        config('key', 'nbdkit_test')
        config('object-size', str(self.obj_size))
        config('size', str(self.dev_size))

        config_complete()
        self.gcs = open(False)

    def tearDown(self) -> None:
        super().tearDown()

    @staticmethod
    def get_data(len):
        buf = secrets.token_bytes(len // 2 + 1)
        return base64.b16encode(buf)[:len]

    def test_zero(self):
        bs = self.obj_size
        ref_buf = bytearray(self.get_data(3 * bs))
        pwrite(self.gcs, ref_buf, 0, 0)

        zero_start = bs // 2
        zero(self.gcs, 2 * bs, zero_start, 0)
        ref_buf[zero_start:zero_start + 2 * bs] = bytearray(2 * bs)

        buf = bytearray(len(ref_buf))
        pread(self.gcs, buf, 0, 0)
        self.assertEqual(buf, ref_buf)

    def test_trim(self):
        ref_buf = bytearray(self.get_data(3 * self.obj_size))
        pwrite(self.gcs, ref_buf, 0, 0)

        obj_count1 = len(list(self.gcs._list_objects(cfg.key_name)))

        trim_start = self.obj_size // 2
        trim(self.gcs, 2 * self.obj_size, trim_start, 0)

        obj_count2 = len(list(self.gcs._list_objects(cfg.key_name)))
        self.assertEqual(obj_count1 - 1, obj_count2)

    def test_readwrite(self):
        ref_buf = self.get_data(2 * self.obj_size)
        start_off = self.obj_size // 2
        pwrite(self.gcs, ref_buf, start_off, 0)
        buf = bytearray(len(ref_buf))
        pread(self.gcs, buf, start_off, 0)
        self.assertEqual(buf, ref_buf)



#!../nbdkit python

disk = bytearray(1024 * 1024)


def open(readonly):
    print("open: readonly=%d" % readonly)
    return 1


def get_size(h):
    return len(disk)


def pread(h, buf, offset):
    end = offset + len(buf)
    buf[:] = disk[offset:end]



#!/usr/bin/env python3
"""
The real nbdkit module is only available when the Python interpreter
runs inside the nbdkit binary. To get unit tests to pass, we provide this
"stub" module that provides just the minimum attributes to do
unit testing.
"""

import logging

log = logging.getLogger(__name__)


FLAG_MAY_TRIM = 1


def parse_size(v):
    return int(v)


def debug(msg):
    log.debug(msg)


def set_error(err):
    pass



# -*- python -*-
# nbdkit
# Copyright Infrascale Inc.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# * Neither the name of Infrascale nor the names of its contributors may be
# used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY INFRASCALE AND CONTRIBUTORS ''AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INFRASCALE OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

# This fake google-cloud-storage module is used to test the gcs plugin.  See
# also tests/test-gcs.sh

import nbdkit


nbdkit.debug("loaded mocked google.cloud.storage module for tests")

buf = b'x'*4096 + b'y'*2048 + b'z'*2048


class Client(object):
    def __init__(self):
        pass

    @staticmethod
    def from_service_account_json(json_credentials_path: str):
        return Client()

    def bucket(self, bucket_name: str):
        return Bucket(self, bucket_name)


class Bucket(object):
    def __init__(self, client: Client, name):
        self.client = client
        self.name = name

    def get_blob(self, key: str):
        return Blob(self, key)

    def blob(self, key: str):
        return Blob(self, key)


class Blob(object):
    def __init__(self, bucket: Bucket, key: str):
        self.bucket = bucket
        self.key = key
        self.size = len(buf)

    def download_as_bytes(self, start: int, end: int, raw_download=True,
                          checksum=None):
        if start is None and end is None:
            b = buf
        elif start is None and end is not None:
            b = buf[:end + 1]
        elif start is not None and end is None:
            b = buf[start:]
        else:
            b = buf[start:end + 1]

        return b



# Example Python plugin.
#
# This example can be freely used for any purpose.

# Run it from the build directory like this:
#
#   ./nbdkit -f -v python ./plugins/python/examples/ramdisk.py test1=foo
#
# Or run it after installing nbdkit like this:
#
#   nbdkit -f -v python ./plugins/python/examples/ramdisk.py test1=foo
#
# The -f -v arguments are optional.  They cause the server to stay in
# the foreground and print debugging, which is useful when testing.
#
# You can connect to the server using guestfish or qemu, eg:
#
#   guestfish --format=raw -a nbd://localhost
#   ><fs> run
#   ><fs> part-disk /dev/sda mbr
#   ><fs> mkfs ext2 /dev/sda1
#   ><fs> list-filesystems
#   ><fs> mount /dev/sda1 /
#   ><fs> [etc]

import nbdkit
import errno

# This is the string used to store the emulated disk (initially all
# zero bytes).  There is one disk per nbdkit instance, so if you
# reconnect to the same server you should see the same disk.  You
# could also put this into the handle, so there would be a fresh disk
# per handle.
disk = bytearray(1024 * 1024)


# There are several variants of the API.  nbdkit will use this
# constant to determine which one you want to use.  This is the latest
# version at the time this example was written.
API_VERSION = 2


# This just prints the extra command line parameters, but real plugins
# should parse them and reject any unknown parameters.
def config(key, value):
    nbdkit.debug("ignored parameter %s=%s" % (key, value))


def open(readonly):
    nbdkit.debug("open: readonly=%d, tls=%r" % (readonly, nbdkit.is_tls()))

    # You can return any Python object from open (even None), and the
    # same object will be passed as the first arg 'h' to the other
    # callbacks below.  To return an error from this method you must
    # throw an exception.
    return 1


def get_size(h):
    return len(disk)


def pread(h, buf, offset, flags):
    end = offset + len(buf)
    buf[:] = disk[offset:end]
    #  or if reading from a file you can use:
    # f.readinto(buf)


def pwrite(h, buf, offset, flags):
    end = offset + len(buf)
    disk[offset:end] = buf


def zero(h, count, offset, flags):
    if flags & nbdkit.FLAG_MAY_TRIM:
        disk[offset:offset+count] = bytearray(count)
    else:
        nbdkit.set_error(errno.EOPNOTSUPP)
        raise Exception



# nbdkit
# Copyright Red Hat
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# * Neither the name of Red Hat nor the names of its contributors may be
# used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY RED HAT AND CONTRIBUTORS ''AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL RED HAT OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

# Used in conjunction with test-timeout.sh

import select
import socket
import sys
import time

unixsocket = sys.argv[1]

sockets = []
t1 = time.monotonic()
for i in range(5):
    s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    print("test-timeout: opened socket %d" % s.fileno(),
          file=sys.stderr, flush=True)
    s.connect(unixsocket)
    # Read and discard the initial greeting from nbdkit.
    b = s.recv(18)
    sockets.append(s)

while sockets:
    ready = select.select(sockets, [], [], 30)[0]
    for s in ready:
        b = s.recv(1024)
        if not b:
            t2 = time.monotonic()
            print("test-timeout: closed socket %d after %ds" %
                  (s.fileno(), t2-t1),
                  file=sys.stderr, flush=True)
            sockets.remove(s)
            s.close()

# At least 5 seconds must have passed before all sockets closed.
t2 = time.monotonic()
print("test-timeout: total elapsed time: %ds" % (t2-t1),
      file=sys.stderr, flush=True)
assert t2-t1 >= 5



# -*- python -*-
# nbdkit
# Copyright Red Hat
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# * Neither the name of Red Hat nor the names of its contributors may be
# used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY RED HAT AND CONTRIBUTORS ''AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL RED HAT OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

# This fake boto3 module is used to test the S3 plugin.  See also
# tests/test-S3.sh

import re


buf = b'x'*4096 + b'y'*2048 + b'z'*2048


class client(object):
    def __init__(
            self,
            type=None,
            aws_access_key_id=None,
            aws_secret_access_key=None,
            aws_session_token=None,
            endpoint_url=None):
        assert type == "s3"
        assert aws_access_key_id == "TEST_ACCESS_KEY"
        assert aws_secret_access_key == "TEST_SECRET_KEY"
        assert aws_session_token == "TEST_SESSION_TOKEN"
        assert endpoint_url == "TEST_ENDPOINT"

    def get_object(
            self,
            Bucket=None,
            Key=None,
            Range=None):
        assert Bucket == "MY_FILES"
        assert Key == "MY_KEY"

        if Range is None:
            # Return the size in a 'ResponseMetadata'.
            return {'ResponseMetadata':
                    {'HTTPHeaders':
                     {'content-length': len(buf)}}}
        else:
            # Range must be something like "bytes=N-M".
            r = re.match("bytes=(\\d+)-(\\d+)", Range)
            assert r
            start = int(r.group(1))
            end = int(r.group(2))
            b = buf[start:end+1]

            # Return the data in a 'StreamingBody'.
            return {'Body': StreamingBody(b)}


class StreamingBody(object):
    def __init__(self, data):
        self._data = data

    def read(self, size=None):
        if size:
            return self._data[:size]
        else:
            return self._data

    def close(self):
        pass
